// ******************************************************
// Functions for atlas specification files
// MRFSEG + GAMIXTURE software bundle: Voxel classification in volumetric 
// images using genetic algorithms and Markov random fields.
// The genetic algorithm is described in: 
// [1]  J. Tohka , E. Krestyannikov, I.D. Dinov , A. MacKenzie-Graham,
// D.W. Shattuck , U. Ruotsalainen, and A.W. Toga. 
// Genetic algorithms for finite mixture model based voxel classification
// in neuroimaging.  
// IEEE Transactions on Medical Imaging  , 26(5):696 - 711, 2007.
// (C) Jussi Tohka, 2005 - 2007 
// Laboratory of NeuroImaging, Department of Neurology, UCLA Medical School, CA, USA
// Institute of Signal Processing, Tampere University of Technology, Finland  
// Version 1.1
// e-mail jussi.tohka@tut.fi
// *****************************************************
// ******************************************************
// Permission to use, copy, modify, and distribute this software 
// for any purpose and without fee is hereby
// granted, provided that the above copyright notice appears in all
// copies.  The author, Tampere University of Technology and University of
// California, Los Angeles make no representations
// about the suitability of this software for any purpose.  It is
// provided "as is" without express or implied warranty.
// *****************************************************



#include "atlasspec.h"

//////////////////////////////////
// the atlas file should look like this:
// lines starting with # are for comments
//////////////////////////////////
// ATLAS NAME
// n numberOfLabels
// Region1name Region1filename LabelNOTPermitted1 2 3 ...
// Region2name Region2filename LabelNOTPermitted1 2 3 ...
//  ....
// Regionnname Regionnfilename
// Label1name Label1type1 Label1type2 Label1type3
// ...
// Labelnname Labelntype1 ....  
// Mrf matrix (numberofLabels + 1) x (numberOfLabels + 1)
// should be symmetric



bool readAtlasSpec(AtlasSpec* atlas,char* filename) 
{
  char str[256];
  char* pstr = str;
  int i,j;
  char c;
  int tmp[256];
  bool probLimits;

  
  ifstream ifile;
  ifile.open(filename);
  if(!ifile.is_open()) return(false);
  ifile >> pstr;
  if(!(strcmp(pstr,"p"))) {
    probLimits = true;
  }
  else { 
    probLimits = false;   
  }
  ifile >> atlas->n;
  ifile >> atlas->numberOfLabels;
 
  if(!probLimits) {
    atlas->regionLowProb = NULL;
    atlas->regionUpProb = NULL;
  }
  else {
    if(atlas->n > 0) {
      atlas->regionLowProb = new float*[atlas->n];
      atlas->regionUpProb  = new float*[atlas->n];
      for(i = 0;i < atlas->n;i++) {
        atlas->regionLowProb[i] = new float[atlas->numberOfLabels];
        atlas->regionUpProb[i]  = new float[atlas->numberOfLabels];
      }
    }
    else { 
      atlas->regionLowProb = new float*[1];
      atlas->regionUpProb  = new float*[1];
      atlas->regionLowProb[0] = new float[atlas->numberOfLabels];
      atlas->regionUpProb[0]  = new float[atlas->numberOfLabels];
    }
  }   

  // reserving the memory 
  if(atlas->n > 0) {
    atlas->filenames = new char*[atlas->n];
    atlas->regionnames = new char*[atlas->n]; 
    atlas->permittedLabels = new LabelList[atlas->n];  
  }

  atlas->labelnames = new char*[atlas->numberOfLabels];
  atlas->labelTypes = new LabelType[atlas->numberOfLabels];
  atlas->mrfConstants = new float*[atlas->numberOfLabels];
  

  for(i = 0;i < atlas->n;i++) {
    atlas->regionnames[i] = new char[MAX_NAME_LEN];
    atlas->filenames[i] = new char[MAX_NAME_LEN];
  } 

  for(i = 0;i < atlas->numberOfLabels;i++) {
    atlas->labelnames[i] = new char[MAX_NAME_LEN];
    atlas->mrfConstants[i] = new float[atlas->numberOfLabels];
   }

  // set the background label info
  strcpy(atlas->labelnames[0],"background");
  atlas->labelTypes[0].pureLabel = true; // pure label;  

  if(probLimits && atlas->n == 0 ) {
    atlas->regionLowProb[0][0] = 0.0;
    atlas->regionUpProb[0][0] = 0.0;
    for(j = 0; j < (atlas->numberOfLabels - 1);j++) {
      ifile >> atlas->regionLowProb[0][j + 1];
      if(atlas->regionLowProb[0][j + 1] < 0.0) atlas->regionLowProb[0][j + 1] = 0.0;
      ifile >> atlas->regionUpProb[0][j + 1];
      if( atlas->regionUpProb[0][j + 1] > 1.0) atlas->regionUpProb[0][j + 1] = 1.0;
    }
  }
  // read the region info
  for(i = 0;i < atlas->n;i++) {
    ifile >> pstr;
    strcpy(atlas->regionnames[i],pstr);
    c = ifile.get();
    ifile >> pstr;
    strcpy(atlas->filenames[i],pstr); 
    if(probLimits) {
      atlas->regionLowProb[i][0] = 0.0;
      atlas->regionUpProb[i][0] = 0.0;
      for(j = 0; j < (atlas->numberOfLabels - 1);j++) {
	ifile >> atlas->regionLowProb[i][j + 1];
        if(atlas->regionLowProb[i][j + 1] < 0.0) atlas->regionLowProb[i][j + 1] = 0.0;
        ifile >> atlas->regionUpProb[i][j + 1];
        if( atlas->regionUpProb[i][j + 1] > 1.0) atlas->regionUpProb[i][j + 1] = 1.0;
      }
      constructLabelList(&(atlas->permittedLabels[i]),atlas->numberOfLabels - 1,NULL,0); 
    }
    else {
      c = ifile.get();
      j = 0;
      if(c == ' ') {     
        while(c == ' ') {
          ifile >> tmp[j];
          c = ifile.get();   
          j++;
        }
      }      
      constructLabelList(&(atlas->permittedLabels[i]),atlas->numberOfLabels - 1,tmp,j);
    }
  } 
  // read the label info
  for(i = 1;i< atlas->numberOfLabels;i++) {
    ifile >> pstr;
    strcpy(atlas->labelnames[i],pstr);
    ifile >> atlas->labelTypes[i].pureLabel;
    if(ifile.get() != '\n') {
      ifile >> atlas->labelTypes[i].mixed[0];
      ifile >> atlas->labelTypes[i].mixed[1];
    }
  }
  
  // read the mrf matrix
  for(i = 0;i < atlas->numberOfLabels;i++) {
    for(j = 0;j < atlas->numberOfLabels;j++) {
      ifile >> atlas->mrfConstants[i][j];
    }
  }
  
  ifile.close();
  return true;
}

void freeAtlas(AtlasSpec* atlas) 
{
  int i;  

  for(i = 0;i < atlas->n;i++) {
    delete[] atlas->regionnames[i];
    delete[] atlas->filenames[i];
  } 

  for(i = 0;i < atlas->numberOfLabels;i++) {
    delete[] atlas->labelnames[i];
    delete[] atlas->mrfConstants[i];
   }
   // reserving the memory 
  delete[] atlas->labelnames;
  delete[] atlas->filenames;
  delete[] atlas->regionnames; 
  delete[] atlas->labelTypes;
  delete[] atlas->mrfConstants;
  delete[] atlas->permittedLabels;  

}

// reads the images in the atlas and ensures that each
// voxelsums  are equal to the unity
// returns 0 if ok

int readAtlasImages(AtlasSpec* atlas,AnalyzeImage** atlasImages)
{
  int i,j,imgsize;
  int intstatus;
  int dimx,dimy,dimz;
  float probsum;
  AnalyzeImage psum;
  bool filter = false;

  // reserve memory
 // atlasImages = new AnalyzeImage*[atlas->n];

  // read images
  for(i = 0;i < atlas->n;i++) {
   // atlasImages[i] = new AnalyzeImage;
    intstatus = readImage(atlas->filenames[i],atlasImages[i]);
    if(intstatus != 0) return(intstatus + i*100);
  }
  // check that the dimensions of the atlas match
  dimx = atlasImages[0]->header.x_dim;
  dimy = atlasImages[0]->header.y_dim; 
  dimz = atlasImages[0]->header.z_dim;
  for(i = 1;i < atlas->n;i++) {
    if(dimx != atlasImages[i]->header.x_dim) return(5 + i*10);
    if(dimy != atlasImages[i]->header.y_dim) return(5 + i*10);
    if(dimz != atlasImages[i]->header.z_dim) return(5 + i*10);
  }
  // make sure that every voxel sums to the unity in prob atlas
  imgsize = dimx*dimy*dimz;
  copyImage(atlasImages[0],&psum);

  for(i = 0;i < imgsize;i++) {
    probsum = 0.0;
    for(j = 0;j < atlas->n;j++) {
      probsum = probsum + atlasImages[j]->data[i];
    } 
    psum.data[i] = probsum;
    if(fabs(probsum) > 0.0001) {
      for(j = 0;j < atlas->n;j++) {
        atlasImages[j]->data[i] =  atlasImages[j]->data[i]/probsum;
      }
    }       
  }
  if(filter) {
   for(i = 0;i < atlas->n;i++) {
      averageFilterZeros(atlasImages[i],&psum,0.0001,1);
    }
    for(i = 0;i < imgsize;i++) {
      probsum = 0.0;
      for(j = 0;j < atlas->n;j++) {
        probsum = probsum + atlasImages[j]->data[i];
      } 
   
      if(fabs(probsum) > 0.0001) {
        for(j = 0;j < atlas->n;j++) {
          atlasImages[j]->data[i] =  atlasImages[j]->data[i]/probsum;
        }
      }       
    } 
  }   
  freeImage(&psum);
  return(0);
}

// releases memory allocated by readAtlasImages

void freeAtlasImages(AtlasSpec* atlas,AnalyzeImage** atlasImages)
{
  int i;

  for(i = 0;i < atlas->n;i++) {
    delete[] atlasImages[i]->data;
  }
}

// takes a mask and masks the atlas (makes atlas values out of mask 0)
// in addition tries to ensure that each voxel within the mask is represented in the atlas
// mask should have value 0 for background and 1 for foreground (i.e. brain) 
// atlas images should be normalized to sum to the unity.

 
bool maskAtlas(AtlasSpec* atlas, AnalyzeImage** atlasImages,AnalyzeImage* mask)
{
    AnalyzeImage probmask;
    int i,j,k,l,ci,cj,ck,imgsize;
    
    // cout << "Entering the maskatlas" << endl;
    // cout << atlasImages[0]->header.x_dim << endl; 
    // cout <<  mask->header.x_dim << endl;
    // cout << atlasImages[0]->header.y_dim << " " <<  mask->header.y_dim << endl;
    // cout << atlasImages[0]->header.z_dim << " " <<  mask->header.z_dim << endl;
   // check that atlas images correspond to the mask image
   if((atlasImages[0]->header.x_dim != mask->header.x_dim) ||(atlasImages[0]->header.y_dim != mask->header.y_dim) 
       || (atlasImages[0]->header.z_dim != mask->header.z_dim)) {
     if((atlasImages[0]->header.x_dim < mask->header.x_dim) ||(atlasImages[0]->header.y_dim < mask->header.y_dim) 
	|| (atlasImages[0]->header.z_dim <  mask->header.z_dim)) {
       cout << "ERROR: Atlas dimensions do not match to the maskfile dimensions" << endl;
       return(false);
     }
     else {
       cout << "Warning:  Atlas dimensions do not match to the maskfile dimensions -> continuing but check the results" << endl;
     }
   }
   newImage(&probmask,atlasImages[0]);
   cout << "probmask created ok" << endl;
   imgsize = (atlasImages[0]->header.x_dim)*(atlasImages[0]->header.y_dim)*(atlasImages[0]->header.z_dim); 
   for(i = 0;i < imgsize;i++) {
     probmask.data[i] = 0;
     for(j = 0;j < atlas->n;j++) {
       probmask.data[i] = probmask.data[i] + atlasImages[j]->data[i];
     }
   }
   // writeImage("tmpprob",&probmask,true);
   //cout << getVoxelValue(mask,190,32,105) << endl;
   // cout << getVoxelValue(&probmask,190,32,105) << endl;

   cout << "probmask generated" << endl;
   for(i = 0;i < mask->header.x_dim;i++ ) {
     for(j = 0;j < mask->header.y_dim;j++ ) {
       for(k = 0;k < mask->header.z_dim;k++ ) {
         // first check if the atlas spatially cover the mask
         if((getVoxelValue(mask,i,j,k) > 0.5) && (getVoxelValue(&probmask,i,j,k) < 0.5) ) { 
	   //  cout << "*";
           if(findClosestNonZero(&probmask,i,j,k,&ci,&cj,&ck,0.5) == false) {
	     cout << "here" << i  << " " << j << " " << k<< endl;
             return(false);
           }
           for(l = 0;l < atlas->n;l++) {
              putVoxelValue(atlasImages[l],i,j,k,getVoxelValue(atlasImages[l],ci,cj,ck));
           }
         }
         
         // then remove those parts of the atlas that are not necessary
         if(getVoxelValue(mask,i,j,k) < 0.5) {
           for(l = 0;l < atlas->n;l++) {
              putVoxelValue(atlasImages[l],i,j,k,0.0);
           }
         }
       }
     }
   }
   // cout << "everything ok" << endl;
   freeImage(&probmask);
   return(true);

}
